#include <codec.h>

#define PERE(i) ((i)>>2)-(((i)&3)?0:1)

/*
 * Code le quadtree dans un fichier FBAB.
 * Retourne le nombre de bits cods, -1 en cas d'erreur.
 */
static int code(quadtree * q, FBAB * f, int diff)
{
    assert(q); assert(f);

    int i, 
	ni = q->taille >> 2, // limite entre les noeuds internes et les feuilles
	taille = 0, // Nombre de bits cods
	calcul;

    for ( i=0 ; i < q->taille ; i++ ) {
	// Racine ou Fils non-uniforme
	if ( i == 0 || ! q->tree[PERE(i)].uniforme ) {
	    // Moyenne
	    if ( i == 0 || ((i-1)&3) != 3 ) {
		// Codage diffrentiel
		if ( diff && i != 0 ) {
		    // Calcul de la diffrence entre pre et fils
		    calcul = q->tree[i].moyenne - q->tree[i>>2].moyenne;
		    calcul = ( calcul > 0 ) ? 2*calcul-1 : -2*calcul;
		    // criture
		    if ( ! (taille += fbab_writediff(calcul, f) ) )
			return -1;
		} else {
		// Codage normal
		    if ( ! fbab_write8(q->tree[i].moyenne, f) )
			return -1;
		    taille += 8;
		}
	    }

	    // Noeud interne
	    if ( i < ni ) {
		// Erreur
		if ( ! fbab_writebit(q->tree[i].erreur & 2, f) )
		    return -1;
		if ( ! fbab_writebit(q->tree[i].erreur & 1, f) )
		    return -1;
		taille += 2;

		// Uniforme
		if ( q->tree[i].erreur == 0 ) {
		    if ( ! fbab_writebit(q->tree[i].uniforme, f) )
			return -1;
		    taille++;
		}
	    }
	}
    }

    return taille;
}

/*
 * Code le quadtree dans le fichier de nom ("qtc/" + q->name + ".qtc")
 */
int codec_code(quadtree * q, int diff)
{
    assert(q);

    FILE * file = NULL;
    char * filename = NULL;
    time_t t = time(NULL);
    int pos_compression = 3, // offset pour crire le taux de compression
	taille_compresse = 0, 
			   // 2^n * 2^n * 8 (taille en bits)
	taille_original = (1<<(q->puissance)) * (1<<(q->puissance)) * 8; 
    FBAB fbab;

    // Ouverture du fichier produit
    filename = malloc(strlen(q->name)+9);
    if ( filename == NULL ) {
	fprintf(stderr,"Impossible de gnrer le nom du fichier compress\n");
	return 0;
    }
    strcpy(filename, "qtc/");
    strcat(filename, q->name);
    strcat(filename, ".qtc");

    file = fopen(filename, "w");
    if ( file == NULL ) {
	perror("fopen");
	return 0;
    }

    free(filename);

    // Initialisation de la structure pour les oprations d'criture bit  bit
    fbab.tampon = 0;
    fbab.position = 7;
    fbab.file = file;

    // Magic number
    fprintf(file, "Q%c\n",diff?'2':'1');

    // Commentaires
				// ctime() imprime dj un '\n'
    pos_compression += fprintf(file, "#%s#Compression:",ctime(&t));
    fputs("            \n",file); // espace pour crire le taux de compression

    // Infos
    unsigned char tmp = q->puissance + 1; // mopmo.
    if ( fwrite(& tmp, sizeof(unsigned char), 1, file) != 1 ) {
	perror("fwrite");
	return 0;
    }
    if ( fwrite(& q->colors, sizeof(unsigned char), 1, file) != 1 ) {
	perror("fwrite");
	return 0;
    }

    // Quadtree
    if ( (taille_compresse = code(q,&fbab,diff)) == -1 ) {
	fprintf(stderr,"Codage impossible\n");
	return 0;
    }

    fbab_flush(&fbab);

    // criture du taux de compression
    fseek(file, pos_compression, SEEK_SET);
    fprintf(file, "%4.2f%%", 
	    ((float)taille_compresse/(float)taille_original)*100);
    fclose(file);

    return 1;
}

/*
 * Dcodage...
 */
static int decode(quadtree * q, FBAB * f, int diff)
{
    assert(q); assert(f);

    unsigned short shdata;
    unsigned char data;
    int i, 
	ni = q->taille >> 2; // limite entre les noeuds internes et les feuilles

    for ( i=0 ; i < q->taille ; i++ ) {
	// Racine ou Fils non-uniforme
	if ( i == 0 || ! q->tree[PERE(i)].uniforme ) {
	    // Moyenne
	    if ( i == 0 || ((i-1)&3) != 3 ) { // Lecture normale
		if ( diff && i != 0 ) {
		    // Dcodage diffrentiel
		    fbab_readdiff(&shdata, f);
		    if ( shdata & 1 ) // data impair => diffrence positive
			q->tree[i].moyenne = q->tree[i>>2].moyenne 
			    + ((shdata+1)>>1);
		    else // pair => diffrence ngative
			q->tree[i].moyenne = q->tree[i>>2].moyenne
			    - (shdata>>1);
		} else {
		    // Dcodage normal
		    if ( ! fbab_read8(&data, f) )
			return 0;
		    q->tree[i].moyenne = data;
		}
	    } else { // Interpolation du 4e fils
		q->tree[i].moyenne = ((q->tree[(i>>2)-1].moyenne)<<2) 
		    - q->tree[i-1].moyenne
		    - q->tree[i-2].moyenne
		    - q->tree[i-3].moyenne
		    + q->tree[(i>>2)-1].erreur;
	    }

	    // Noeud interne
	    if ( i < ni ) {
		// Erreur
		if ( ! fbab_readbit(&data, f) )
		    return 0;
		if ( data ) 
		    q->tree[i].erreur |= 2;
		if ( ! fbab_readbit(&data, f) )
		    return 0;
		if ( data )
		    q->tree[i].erreur |= 1;

		// Uniforme
		if ( q->tree[i].erreur == 0 ) {
		    if ( ! fbab_readbit(&data, f) )
			return 0;
		    q->tree[i].uniforme = data;
		}
	    }
	// Fils uniforme
	} else { 
	    q->tree[i].moyenne = q->tree[(i>>2)-((i&3)?0:1)].moyenne;
	    if ( i < ni ) {
		q->tree[i].erreur = q->tree[(i>>2)-((i&3)?0:1)].erreur;
		q->tree[i].uniforme = q->tree[(i>>2)-((i&3)?0:1)].uniforme;
	    }
	}
    }

    return 1;
}

/*
 * Retourne la taille du quadtree (nombre d'lments de l'arbre)
 * calcul  partir de la puissance de 2 qui donne la hauteur/largeur de l'image
 */
static int taille_quadtree(int n)
{
    int add = 1, taille = 0;

    // somme de k=(0  n) de 4^k
    while ( n >= 0 ) {
	taille |= add;
	add <<= 2;
	n--;
    }

    return taille;
}

/*
 * Retourne un quadtree dcod  partir du fichier filename
 */
quadtree * codec_decode(char * filename)
{
    FILE * file = NULL;
    char magic[3];
    int c;
    quadtree * q = NULL;
    FBAB fbab;
    char * temp = NULL;

    // Ouverture
    file = fopen(filename, "r");
    if ( file == NULL ) {
	perror("fopen");
	return NULL;
    }

    // Initialisation de la structure pour les oprations de lecture bit  bit
    fbab.tampon = 0;
    fbab.position = 7;
    fbab.file = file;

    // Vrification du magic number
    fread(magic, sizeof(char), 3, file);

    if ( magic[0] != 'Q' 
	    || magic[2] != '\n' 
	    || ( magic[1] != '1' && magic[1] != '2' ) ) {
	fprintf(stderr,"%s n'est pas un fichier quadtree valide.\n",filename);
	fclose(file);
	return NULL;
    }

    // Zap les commentaires
    while ( (c = getc(file)) == '#' ) {
	while ( getc(file) != '\n' )
	    ;
    }
    ungetc(c,file);

    // Initialisation de mmoire
    q = malloc(sizeof(quadtree));
    if ( q == NULL ) {
	fprintf(stderr,"Impossible d'allouer la mmoire pour rcuprer"
		" le quadtree\n");
	fclose(file);
	return NULL;
    }

    // Lecture des informations
    if ( fread(& q->puissance, sizeof(unsigned char), 1, file) != 1 ) {
	perror("fread");
	return NULL;
    }
    q->puissance--;
    if ( fread(& q->colors, sizeof(unsigned char), 1, file) != 1 ) {
	perror("fread");
	return NULL;
    }

    // Mise en place du nom de l'image
    temp = strdup(filename);
    temp[strlen(temp)-4] = 0; // On retire .qtc
    q->name = strdup(basename(temp));
    free(temp);

    // Allocation de mmoire pour le quadtree
    q->taille = taille_quadtree(q->puissance);
    q->tree = calloc(q->taille, sizeof(qt_element));
    if ( q->tree == NULL ) {
	fprintf(stderr,"Impossible d'allouer la mmoire pour le quadtree\n");
	free(q->name);
	free(q);
	fclose(file);
	return NULL;
    }

    // Dcodage du quadtree
    if ( ! decode(q,&fbab,magic[1]=='2') ) {
	fprintf(stderr,"Dcodage impossible\n");
	fclose(file);
	free(q);
	return NULL;
    }

    fclose(file);

    return q;
}
